
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2007 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.0. Last modification 1/03/07
*)
(* ..................................................................
    Module.......: Types_unify
    Initial Date : 14 Octobre 2003
    Last Update..: 21 Jully 2005
    Developer....: Mathieu Turuani
    Maintenance..: Judson Santiago
    Compiler.....: OCAML 3.07 
    Comments.....: Declaration of all types used in Atse
    Project .....: The CL-Atse Protocol Analyser, AVISPA.
 .................................................................... *)


(* ========================
   Types used by the parser 
   ======================== *)


(* Stores the basic elements of a state *)
type termGL = 
    EAtm of string 
  | EVar of string 
  | Nonce of string
  | Cuple of termGL list 
  | PubCypher of termGL*termGL
  | SymCypher of termGL*termGL 
  | Prim of termGL
  | EXor of termGL list
  | EExp of termGL * (termGL list)
;;


(* Gives are secrets given to the intruder, so they should not be
   considered as secrets. They are used only in IFv1 *)
type givesIF = 
    termGL
;;


(* Stores secret facts. The first field is the secret and the second
   is the official owner *)
type secretsIF = 
    termGL * termGL  
;;


(* Stores the request terms. An agent "A" should do a
   Request(B,A,cNb,Nb) to indicate the acceptance of the value Nb as
   coming from the agent B. For each request it should exist in the past
   a corresponding witness *)
type requestKind = 
    Strong
  | Weak
;;

type requestIF = 
    termGL * termGL * requestKind * termGL  (* WitPrin, ReqPrin, Kind, term *)
;;


(* Stores the witness terms. Witness(B,A,cNb,Nb') means that a fresh
   value Nb' was generated by B and sent to A. "cNb" identifies the
   witness. You need this identification because you can have more than
   one witness for a pair of agents *)
type witnessIF = 
    termGL * termGL * termGL    (* WitPrin, ReqPrin, term *)
;;


(* Stores messages of the protocol *)
type mtermIF = 
    termGL
;;


(* Stores an state of the protocol: a set of agents, keys, nonces,
   etc.  The first argument is a string representing the agent playing
   this step of the role *)
type wtermIF = 
    string * termGL
;;


(* The conditions from the right side of a rule, if they exist, are     *)
(* structured as add a new contains prop. and removing a contains prop. *)
type containsIF =
    ContainsIF of termGL * termGL
  | NotContainsIF of termGL * termGL
;;


(* The conditions from the left side of a rule, if they exist, are      *)
(* structured as an equality, not-equality  and not-contains conditions *)
type conditionIF =
    EqualIF of termGL * termGL
  | NotEqualIF of termGL * termGL
;;


(* The nice IF syntax allows contains facts both in states and conditions. *)
(* We need to deal with that.....					   *)
type generic_conditionIF =
    TrueCondition of conditionIF
  | HiddenContains of containsIF
;;


(* The rules of the protocol are stored in this structure *)
type termIF = 
    wtermIF list * mtermIF list * witnessIF list * requestIF list * secretsIF list * containsIF list
;;


(* The intruder knownledge has this type  *)
type knowIF = 
    termGL list
;;


(* The name of a step in the IF file *)
type nameIF =
    string
;;


(* This type caracterize an step of the protocol  *)
type stepIF =                     (* Equals, etc *)  (* From State *)  (* From Tests *)
    nameIF * wtermIF list * mtermIF list * conditionIF list * containsIF list * containsIF list * termIF
;;


(* This type describe the list of type declarations in the IF.  
   First : atomic term.  Second : type name.  *)
type types_section = 
    (termGL * string) list
;;

(* Stores facts of an IF file  *)
type factIF =
  | State of wtermIF
  | Iknows of mtermIF
  | Witness of witnessIF
  | Request of requestIF
  | Secret of secretsIF
  | ContainFact of containsIF
;;

type attack_state_elem =
  | PositiveSt of factIF
  | NegativeSt of factIF
  | ConditionSt of conditionIF
;;

type attack_state =
    attack_state_elem list
;;

(* This is the type returned by the parser to AtSe *)
type parseReturn =
     stepIF list * (knowIF * wtermIF list * secretsIF list * containsIF list) list * types_section * attack_state list
;;



(* ===============================
   Types used by atse main modules
   =============================== *)

(* Sign type. *)
type sign = Pos | Neg ;;

(* Most bacic element used by Atse *) 
type term = 
  | Uplet of term list 
  | Xor of term list
  | Exp of term * ((term * sign) list)
  | PCrypt of term*term 
  | SCrypt of term*term
  | PInv of term 
  | Atm of int 
  | Var of int
;;

(* Substitution : Var number = Term purified. *)
type t_equal = 
    (int * term) list
;;    

(* Constraint   : Term \neq Term *)
type t_unequal = 
    (term * term) list
;;      

(* Debug only *)
type t_freevar = FreeVar of int;;

(* Subst * free var number * Constraint *)
type t_subst = 
    t_equal * t_freevar * t_unequal
;; 

(* std_ls * xor_eqns * rem_ls * free var number * constraint *)
type t_split_subst = 
    t_equal * term list list * t_equal * t_freevar * t_unequal
;; 

(* Possible results for crtTest function *)
type ctrTestResult = 
  | Equal 
  | NeverEqual 
  | Uncertain
;;

(* In (m,k) we get m if we can build k *)
type t_todec = 
    (term * term) list
;;          

type t_know_elem = 
  | H of int 
  | D of term
;;

type t_know  = 
    t_know_elem list
;;

type t_request  = 
    (requestKind * term) list
;;

type t_witness = 
    term list
;;

(* All the secrets *)
type t_secret = 
    SecrecyOf of (term*term) list
;;                  

type t_sets_elem = 
    InSet of term
;;

type t_sets =
    (term list) * (t_sets_elem list)
;;

(*
type t_lost = 
    Lost of t_secret
;;
*)

type t_etat = 
    t_todec * t_know * t_witness * t_secret * t_sets * t_subst
;;

type t_system = 
    t_etat list
;;

type t_forw = 
  | Unify of term * (term list) 
  | Hypot of term list
;;

(* Table containing all the types found in an IF file. 
   To each type is associated a number *)
type tabletype = 
    (int * string) list
;;

(* Table storing the list of if role instances. *)
type t_role_table =
    (string * (string list)) list
;;

type t_msg = 
    term
;;

type t_TestSet =
    TestInSet of term
  | TestNotInSet of term
;;

type t_NewSet =
    AddToSet of term
  | RemoveFromSet of term
;;

type t_secrecygoal = 
    t_secret
;;

(* Protocol Step. It is : messages received and sent, two terms to be unified (equality condition), 
   a set of secrecy goals, some witness and requests, a list of unequality conditions, a set of
   tests on set elements (TestIn, TestNotIn), and a set of set actions (Add, Remove). *)
type t_step = 
    t_msg * t_msg * term * term * t_secrecygoal * t_witness * t_request * t_unequal * t_TestSet list * t_NewSet list 
;;

(* Kind of protocol step : sooner for a step that can be executed as soon as possible, Maybe_Sooner for a step
   that has a chance to be executable as soon as possible (modulo other conditions that I just don't remember,
   see the main module), and Generic for all ther steps. *)
type t_StepKind =
    Sooner
  | Generic
  | Maybe_Sooner
;;

type t_trace = 
    (string * nameIF * t_step * t_equal) list
;;

(* A protocol role is a tree. Step is a "true" protocol step. "Choice" is a point where one and only
   one of the following list of roles can be executed. This represents the non determinism of the If 
   file, which represents choice points of honest participants. EndRole mark the end. *)
type t_role = 
    Step of t_step * t_StepKind * t_role * t_trace | Choice of t_StepKind * t_role list | EndRole
 ;;

(* A protocol specification is just a list of roles, with the participant's name. *)
type t_specif = 
    (string * t_role) list
;;

type t_state = 
    ( t_system * t_trace * t_specif) list
;;

type t_unforgeable =
    term list

(* the first element of a class is the representative of the class,
   e.g. if the partition is [[1;2;3];[4;5];[6]] then it consists of
   three classes with the representatives 1, 4 and 6. 
   There must not exist an empty class, i.e. [...;[];...] is not allowed. *)
type t_partition = int list list;;


(* Represents a partial order on in the following way:
   if (n,[m1,m2,...,mx]) is in the partial order, this means that 
   n < m1, n < m2, ..., n < mx and that n is not less equal than 
   anything else. 
   "[]" is the empty partial order for the elements in l. *)
type t_po = (int * (int list)) list;;


type t_attack_state =
  | Atk_Iknows    of term
  | Atk_NotIknows of term
  | Atk_Set       of t_TestSet
  | Atk_Equal     of term * term
  | Atk_NotEqual  of term * term
;;

type t_attack_state_goal =
  t_attack_state list
;;

(* The type returned when a flaw is found. 
   It describes the kind of flaw. *)
type atkKind = 
  | Secr of term 
  | Auth of term
  | NRep of t_attack_state list
;;

(* Types, used during unification *)

(* prestate occurs before variable identification and will be transformed into
   a list of prestates *)
type t_unif_prestate = 
    (* std               xor              shared_vars  rem_sub *)
    (term * term) list * term list list * int list * t_subst 
;;

(* a list of states is generated by choose_theory, 
   which takes a list of prestates as input *)
type t_unif_state = 
    (* std             * xor            * vars in std * vars in xor * rem_sub *)
    (term * term) list * term list list * int list    * int list    * t_subst 
;;


type t_vi_ctrs = ((int * int) list) * (t_partition list);;


(* EXCEPTIONS *)

exception Invalid_Var_Identif;;
exception No_Solution;;
